import numpy as np
import matplotlib.pyplot as plt
from tabulate import tabulate
from polyagamma import random_polyagamma
from scipy.stats import multivariate_normal
from sklearn.preprocessing import StandardScaler
from sklearn.utils.validation import check_is_fitted
from .bclr_helper import _proc_mean_cov

inv = np.linalg.inv
det = np.linalg.det
    
class BayesCC:
    """
    This class implements the Bayesian Changepoint via Logistic Regression (bclr) method 
    described in Thomas, Jauch, and Matteson (2023).
    """
    def __init__(self, X, prior_mean, prior_cov, n_iter, 
                 prior_kappa=None, scaled=False, burn_in=None):
        """

        Parameters
        ----------
        X : array-like of shape n x d
            Array of n d-dimensional vectors to assess changepoints.
        prior_mean : ndarray of length d
            Array containing prior mean of Normal distribution.
        prior_cov : ndarray of shape (d,d)
            Symmetric positive (semi)definite prior covariance matrix. 
        n_iter : int
            Number of iterations to run Gibbs sampler. 
        prior_kappa : ndarray of length n-1 nonnegative values.
            Prior distribution for the changepoint kappa. 
            The default is None (i.e. uniform).
        scaled : bool, optional
            If False, each column of data will be mean centered and 
            standardized to have variance 1. The default is False.
        burn_in : int, optional
            Number of initial iterations of the Gibbs sampler to discard. 
            The default is None.

        Raises
        ------
        ValueError
            Data array must be two-dimensional.

        Returns
        -------
        None.

        """
        if len(X.shape) != 2: 
            raise ValueError("Array must be 2-dimensional")
    
        self.n, self.p = X.shape
    
        if not scaled:
            self.std_sc = StandardScaler()
            self.std_sc.fit(X)
            self.X = self.std_sc.transform(X)
        else: 
            self.X = X
        self.scaled = scaled
        
        if n_iter <= 0:
            raise ValueError("Number of MC iterations should be positive")
        
        if burn_in == None:
            burn_in = n_iter/2
        elif burn_in <= 0:
            raise ValueError("burn_in should take only positive values")
        elif burn_in > n_iter:
            raise ValueError("burn_in should be smaller than n_iter")

        self.prior_mean, self.prior_cov = _proc_mean_cov(prior_mean, prior_cov, self.p)
            
        if prior_kappa is None:
            self.prior_kappa = np.repeat(1, self.n-1)
        else:
            if np.any(prior_kappa < 0):
                raise ValueError("Must have nonnegative weights for prior.")
            elif len(prior_kappa) != (self.n-1):
                raise IndexError("Length of prior distribution must be as long as number of potential changepoints")
            self.prior_kappa = prior_kappa
            
        self.n_iter = n_iter
        self.burn_in = int(burn_in)
    
    def fit(self, init_k = None, init_beta = None, tol = 1e-12, c = 1e-2, rng = None):
        """
        Fit BayesCC class, meaning implement the Gibbs sampler discussed for drawing posterior changepoints and coefficients in 
        Thomas, Jauch, and Matteson (2025).

        Parameters
        ----------
        init_k : int, optional
            The initial changepoint estimate to start the Markov chain. The default is None, in which case it is set
            to be equal to the midpoint of the sequence.
        init_beta : array_like of shape (d,) (where d is dimensionality of the data), optional
            Initial value of the coefficient vector beta. The default is None, in which case it is set to be equal 
            to the all zero coefficient vector.
        tol : float, optional
            Value below which to ignore kappa prior probabiltiies. Should be positive.
            The default is 1e-12.
        c : float, optional
            Value at which to set the smallest possible non-zero 'probability'. Should be positive.
            The default is 1e-2. 
        rng : np.random._generator.Generator, optional
            Random number generator to ensure reproducibility. The default is None.
            
        Returns
        -------
        None.

        """
        
        if rng == None:
            rng = np.random.default_rng()
        
        if init_k == None:
            init_k = self.n/2
        
        if init_beta is None:
            init_beta = np.repeat(0, self.p)
        
        self.k_draws_ = np.empty(self.n_iter)
        self.k_draws_[0] = int(init_k)
        
        self.beta_draws_ = np.empty((self.n_iter, self.p))
        self.beta_draws_[0] = init_beta
        
        self.omega_draws_ = np.zeros((self.n_iter, self.n))
        
        for t in range(1, self.n_iter):
            zv = np.squeeze(self.X @ self.beta_draws_[t-1,None].T)
            self.omega_draws_[t] = random_polyagamma(h=1, z=zv, size=self.n, random_state=rng)
            
            inv_cov = inv(self.prior_cov)
            V_omega = inv(self.X.T @ np.diag(self.omega_draws_[t]) @ self.X + inv_cov)
            kappa = np.expand_dims(np.concatenate([np.repeat(-1/2, self.k_draws_[t-1]), np.repeat(1/2, self.n-self.k_draws_[t-1])]), axis=1)
            m_omega = np.squeeze(V_omega @ (self.X.T @ kappa + inv_cov @ np.expand_dims(self.prior_mean, 1)))
            self.beta_draws_[t] = multivariate_normal.rvs(mean=m_omega, cov=V_omega, size=1, random_state=rng)
                                                                                                                                                                          
            k_lpvec = np.zeros(self.n-1)
            logk_lpvec = np.zeros(self.n-1)
            prior_kappa_pos = np.where(self.prior_kappa > tol)[0]
            
            logk_lpvec[prior_kappa_pos[0]] = np.log(c)
            for k in range(1,len(prior_kappa_pos)):
                k1 = prior_kappa_pos[k]
                k0 = prior_kappa_pos[k-1]
                logk_lpvec[k1] = logk_lpvec[k0] - (self.X[k1,:] @ self.beta_draws_[t,None].T)[0] + np.log(self.prior_kappa[k1]) - np.log(self.prior_kappa[k0])
            
            v = np.max(logk_lpvec)
            # to avoid overflow, we truncate log posterior kappa values to 100
            if v > 100:
                logk_lpvec = logk_lpvec - v + 100
                
            k_lpvec = np.exp(logk_lpvec)
            k_pvec = k_lpvec/np.sum(k_lpvec)
            
            self.k_draws_[t] = rng.choice(np.arange(1, self.n), size=1, p=k_pvec)[0]

            
        self.post_k = self.k_draws_[self.burn_in:]
        self.post_beta = self.beta_draws_[self.burn_in:, :]
            
    def transform(self, verbose=True):
        """
        Calculate the draws of kappa and beta outside of the burn-in period and calculate probabilities. 
        Optional displays a table with estimates of kappa and corresponding probabilities.

        Parameters
        ----------
        verbose : bool, optional
            Whether to print out a table of posterior changepoint estimates. The default is True.
        beta_original: bool, optional
            Whether or not to produces beta terms on the original scale. The default is False.
            
        Returns
        -------
        None.

        """
        check_is_fitted(self)
        #Here we create the values outside of the burn_in and calculate probabilities
        post_k_vals, post_k_counts = np.unique(self.post_k, return_counts=True)
        
        arr0 = post_k_counts/(self.n_iter-self.burn_in)
        arr = arr0[arr0 > 0]
        if np.array_equal(arr, np.array([1.])):
            self.norm_entr = 0
        else:
            self.norm_entr = np.sum(-arr * np.log(arr))/np.log(self.n-1)
            
        self.post_k_mode = post_k_vals[np.argmax(post_k_counts)]
        self.post_mode_prob = np.max(arr)
        self.post_beta_mean = np.mean(self.post_beta, axis=0)
        self.post_beta_snr = (self.post_beta_mean)**2/np.var(self.post_beta, axis=0)
        
        if verbose:
            table = [post_k_vals, post_k_counts/(self.n_iter-self.burn_in)]
            print(tabulate(table, headers='firstrow', tablefmt='fancy_grid'))
        
    def plot_k(self):
        """
        Plots the posterior probability mass function of the changepoints.

        Returns
        -------
        None.

        """
        check_is_fitted(self)
        bins = np.arange(1, self.n)
        plt.hist(self.post_k, rwidth=0.9, density=True, align='left', bins=bins)
        plt.ylabel("Posterior probability")
        plt.xlabel(r"$\kappa$")
        plt.show()
        
    




import numpy as np
import pandas as pd
from .bclr_one import BayesCC
from .bclr_helper import uni_binom, prob_mode, SegmentationWarning, _proc_cov
import warnings

inv = np.linalg.inv
det = np.linalg.det

class MultiBayesCC:
    """
    Here we utilize an off-the-shelf changepoint method and post-process using bclr for better understanding of changepoint
    uncertainty and the best representation of said changepoint. This uses the dynamic programming implementation in the ruptures package.

    """
    def __init__(self, X, cps, prior_cov, n_iter=1000, lam=0, min_size=10, rng = None, warnings=True):
        """
        
        Parameters
        ----------
        X : array-like of shape n x d
            Array of n d-dimensional vectors to assess changepoints.
        cps : int or list
            Number of changepoints to seed or list of initial changepoints. 
            If list of changepoints given, note that it should be in terms of the indices
            series {1, 2, ..., len(X)}.
        prior_cov : ndarray of shape (d,d)
            Symmetric positive (semi)definite covariance matrix, 
            for each segment of series.
        n_iter : int
            Number of iterations to run Gibbs sampler for each segment.
        lam : float
            Interpolation parameter for "uni-binomial" prior. Between 0 and 1.
        min_size : int, optional
            Minimum distance between changepoints. The default is 10.
        rng : np.random._generator.Generator, optional
            Random number generator to ensure reproducibility. The default is None.
        warnings : bool, optional
            If False, suppress warnings. 

        Returns
        -------
        None.

        """
        if rng == None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng
        
        if not type(self.rng) == np.random._generator.Generator:
            raise TypeError("rng should be of type np.random._generator.Generator")
        
        tcps  = type(cps)
        if not (tcps == int or tcps == list):
            raise TypeError("cps should be an integer or a list of changepoint locations")
        
        if tcps == int:
            self.bkps = list(np.linspace(0, len(X), cps+2, dtype=np.int64))
        else:
            self.bkps = [0] + cps + [len(X)]              
        
        self.K = len(self.bkps)-2
        self.n, self.p = X.shape
        if np.floor(self.n/(2*self.K + 2)) <= min_size:
            raise ValueError("min_size is too large for any changepoints to be estimated")
            
        if n_iter <= 0:
            raise ValueError("Number of MC iterations should be positive")
        self.prior_cov = _proc_cov(prior_cov, self.p)
        self.n_iter = n_iter
        
        if lam < 0 or lam > 1:
            raise ValueError("Lambda must be nonnegative and no greater than 1")
        else:
            self.lam = lam
        
        self.X = X
        self.transformed = False
        
        if min_size <= 1:
            raise ValueError("min_size should be greater than 1")

        self.min_size = int(min_size)
        self.warnings = warnings
        
    def fit(self):
        """
        Fit MultiBayesCC class, meaning implement the Gibbs sampler on each consecutive segment, 
        according to the multiple changepoint formulation discussed for drawing posterior changepoints 
        and coefficients in Thomas, Jauch, and Matteson (2025).
    
        Returns
        -------
        None.

        """
        prior_mean = np.repeat(0, self.p)
        self.transformed = False
        self.prior_kappas = [uni_binom(n=self.bkps[i+2]-self.bkps[i]-1, 
                                       p=(self.bkps[i+1]-self.bkps[i])/(self.bkps[i+2]-self.bkps[i]), 
                                       lam=self.lam) for i in range(self.K)]
        self.bccs_ = [BayesCC(self.X[self.bkps[i]:self.bkps[i+2], :], prior_mean, 
                             self.prior_cov, self.n_iter, prior_kappa=self.prior_kappas[i])
                     for i in range(self.K)]
        for i in range(self.K):
            self.bccs_[i].fit(rng = self.rng)
    
    def transform(self):
        """
        Calculate posterior distributions and summaries for each segment defined by the 
        endpoints of ``bkps`` defined above. 

        Returns
        -------
        None.

        """
        if self.transformed:
            pass
        else:
            for i in range(self.K):
                self.bccs_[i].post_k = self.bccs_[i].post_k+self.bkps[i]
                self.bccs_[i].transform(verbose=False)
        
        self.transformed = True
    
    def proc(self):
        self.fit()
        self.transform()
        self.transformed = False

    def cps_df(self, thr=None, offset=0):
        """
        Returns estimated changepoints (posterior modes), along with posterior 'probability', and normalized entropy. 
        Values coded nan (removed due to  ``min_size`` constraints are not returned). 

        Parameters
        ----------
        offset : float, optional
            For use in time series with indexing representing time: e.g. years, etc. Adds this
            value to the estimated changepoints. The default is 0.
        thr : float, optional
            Remove changepoints from DataFrame with normalized entropy greater than threshold.
            Between 0 and 1. The default is None.

        Returns
        -------
        df : pandas.DataFrame
            DataFrame with the above described information. 

        """
        
        bc_info = []
        prev = 0
        for i, bc in enumerate(self.bccs_):
            if i >= 1:
                pot = bc_info[i-1][0]
                if np.isnan(pot):
                    pass
                else:
                    prev = pot
                mode_val = prob_mode(bc.post_k[np.logical_and(bc.post_k >= prev-offset + self.min_size, bc.post_k <= self.bkps[i+2]-self.min_size)])
                bc_info.append((mode_val+offset, bc.post_mode_prob, bc.norm_entr))
            else:
                bc_info.append((bc.post_k_mode+offset, bc.post_mode_prob, bc.norm_entr))
        
        df = pd.DataFrame(bc_info, columns = ['Location', 'Posterior Probability', 'Normalized Entropy'])
        df_red = df[df['Location'].notnull()]
        if thr is None:
            return df_red
        else:
            if thr < 0 or thr > 1:
                raise ValueError("Threshold thr should be between 0 and 1 (inclusive)")
            else:
                return df_red[df_red['Normalized Entropy'] < thr]
    
    def fit_predict(self, iter_sch = [100, 250], thr_sch = [0.75, 0.5], offset=0):
        """
        Predict changepoints after two successive warm-up periods of increasing "complexity".

        Parameters
        ----------
        iter_sch : list of increasing positive int, optional
            List of increasing number of iterations to run Gibbs sampler in first two warm-up periods. 
            The default is [100, 250].
        thr_sch : List of decreasing float between 0 and 1, optional
            List of decreasing entropy thresholds. The default is [0.75, 0.5].

        Returns
        -------
        pd.DataFrame
            Estimated changepoints, posterior probability and normalized entropy.

        """
        if len(iter_sch) != 2 or len(thr_sch) != 2:
            raise ValueError("Please ensure iter_sch and thr_sch are both of length 2")
        
        self.warm_up(n_iter_w=iter_sch[0], thr=thr_sch[0])
        self.warm_up(n_iter_w=iter_sch[1], thr=thr_sch[1])
        self.fit()
        self.transform()
        return self.cps_df(offset=offset)
    
    def fit_transform(self):
        """
        Fits, then transforms.

        Returns
        -------
        None.

        """
        self.fit()
        self.transform()
    
    def warm_up(self, n_iter_w=100, thr=None):
        """
        Runs the chain with various initializations according to Section 6.1 of Thomas, Jauch, and Matteson (2025).
        Argument thr can be useful for enhanced estimation ability by removing spurious changepoints. Resets ``bkps`` and
        may reduce number of changepoints ``J``. 

        Parameters
        ----------
        n_iter_w : int, optional
            Number of iterations to run each "warm-up" chain. The default is 100 and the minimum value is 50.
        thr : float, optional
            Remove changepoints with normalized entropy greater than threshold.
            Between 0 and 1. The default is None.

        Returns
        -------
        None.

        """
        n_iter_w = int(max(50, n_iter_w))
        self.n_iter_init = self.n_iter
        self.n_iter = n_iter_w
            
        self.proc()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            dfM = self.cps_df(thr)['Location']
        dfM_nan = dfM[np.logical_not(np.isnan(dfM))]
        self.bkps = [0] + list(dfM_nan.astype(np.int32)) + [self.n]
        if len(dfM_nan) < self.K:
            if self.warnings:
                warnings.warn_explicit(message="""Number of changepoints reduced due to nan values owing to min_size constraints specified in MultiBayesCC... \n""",
                          category=SegmentationWarning, filename="bclr_multi.py", lineno=246)
            self.K = len(dfM_nan)
        
        self.n_iter = self.n_iter_init
        
        
        



from ..bclr_multi import MultiBayesCC
import pytest
import numpy as np

Xm = np.load("bclr_multi.npy")

prior_covBAD1 = np.array([[3, 1], [2,2]])
prior_covBAD2 = np.array([[1, 1], [1,0.999]])
prior_covBAD3 = np.array([[1,0,0], [0,1,0], [0,0,1]])
prior_covBAD4 = np.array([[1,0], [1,0], [0,1]])
prior_mean = [0,0]
prior_cov = np.array([[1,0], [0,1]])

@pytest.mark.parametrize('prior_covBAD', [prior_covBAD1, prior_covBAD2, prior_covBAD3, prior_covBAD4])
def test_bad_cov(prior_covBAD):
    with pytest.raises(ValueError) as e_bad_cov:
        MultiBayesCC(Xm, cps=2, prior_cov = prior_covBAD, n_iter=1000)
    
    assert e_bad_cov.type is ValueError

def test_fit_transform_cps():
    bclr_multi_good = MultiBayesCC(Xm, cps=2, prior_cov = prior_cov, n_iter=1000)
    bclr_multi_good.fit_transform()
    bclr_multi_good.transform() #check additional transform does nothing...
    est_cps = bclr_multi_good.cps_df()['Location'].to_numpy()
    np.testing.assert_almost_equal(est_cps, np.array([200, 500]), 2)
   
@pytest.mark.filterwarnings("ignore:Number of changepoints reduced due to nan values")
@pytest.mark.parametrize('warn,cps', [(True, 10), (False, [100, 250, 550])])
def test_fit_predict_cps(warn,cps):
    bclr_multi_good = MultiBayesCC(Xm, cps=cps, prior_cov = prior_cov, n_iter=1000, warnings=warn)
    est_cps = bclr_multi_good.fit_predict()['Location'].to_numpy()
    np.testing.assert_almost_equal(est_cps, np.array([200, 500]), 2)
  
@pytest.mark.filterwarnings("ignore:Number of changepoints reduced due to nan values")
@pytest.mark.parametrize('warn', [True, False])
def test_warm_up_cps(warn):
    bclr_multi_good = MultiBayesCC(Xm, cps=5, prior_cov = prior_cov, n_iter=1000, warnings=warn)
    bclr_multi_good.warm_up(n_iter_w = 400, thr=0.8)
    bclr_multi_good.fit_transform()
    est_cps = bclr_multi_good.cps_df(thr=0.8)['Location'].to_numpy()
    np.testing.assert_almost_equal(est_cps, np.array([200, 500]), 2)
    
@pytest.mark.parametrize("index", [1,2,3,4,5,6])
def test_init(index):
    if index==1:
        with pytest.raises(TypeError):
            MultiBayesCC(Xm, cps=2, prior_cov = prior_cov, n_iter=1000, warnings=False, rng=100)
        
    elif index==2:
        with pytest.raises(TypeError):
            MultiBayesCC(Xm, cps=3.5, prior_cov = prior_cov, n_iter=1000, warnings=False)
        
    elif index==3:
        with pytest.raises(ValueError):
            MultiBayesCC(Xm, cps=2, prior_cov = prior_cov, n_iter=1000, warnings=False, min_size=200)
        
    elif index==4:
        with pytest.raises(ValueError):
            MultiBayesCC(Xm, cps=2, prior_cov = prior_cov, n_iter=1000, warnings=False, min_size=1)
    
    elif index==5:
        with pytest.raises(ValueError):
            MultiBayesCC(Xm, cps=2, prior_cov = prior_cov, n_iter=0, warnings=False)
        
    else:
        with pytest.raises(ValueError):
            MultiBayesCC(Xm, cps=2, prior_cov = prior_cov, n_iter=1000, warnings=False, lam=2)

@pytest.mark.parametrize("iter_sch,thr_sch", [[[100, 250], [1.5, 0.5]], [[100, 250, 500],[0.75, 0.5]]])
def test_bad_fit_predict(iter_sch, thr_sch):
    with pytest.raises(ValueError):
        bclr_multi_bad = MultiBayesCC(Xm, cps=2, prior_cov = prior_cov, n_iter=1000, warnings=False)
        bclr_multi_bad.fit_predict(iter_sch=iter_sch, thr_sch=thr_sch)


# __init__.py
from .bclr_one import BayesCC 
from .bclr_helper import std_video, gen_sim, uni_binom
from ._version import __version__
from .bclr_multi import MultiBayesCC

__all__ = ['BayesCC', 'MultiBayesCC', 'uni_binom', 'std_video', 
'gen_sim', '__version__']



from ..bclr_one import BayesCC
import pytest
import numpy as np

X = np.load("bclr_one.npy")

prior_meanBAD1 = [0]
prior_meanBAD2 = [[3,4], [1,2]]
prior_covBAD1 = np.array([[3, 1], [2,2]])
prior_covBAD2 = np.array([[1, 1], [1,0.999]])
prior_covBAD3 = np.array([[1,0,0], [0,1,0], [0,0,1]])
prior_covBAD4 = np.array([[1,0], [1,0], [0,1]])
prior_mean = [0,0]
prior_cov = np.array([[1,0], [0,1]])

@pytest.mark.parametrize('prior_meanBAD', [prior_meanBAD1, prior_meanBAD2])
def test_bad_mean(prior_meanBAD):
    with pytest.raises(ValueError) as e_bad_mean:
        BayesCC(X, prior_mean = prior_meanBAD, prior_cov = prior_cov, n_iter=1000)
    
    assert e_bad_mean.type is ValueError

@pytest.mark.parametrize('prior_covBAD', [prior_covBAD1, prior_covBAD2, prior_covBAD3, prior_covBAD4])
def test_bad_cov(prior_covBAD):
    with pytest.raises(ValueError) as e_bad_cov:
        BayesCC(X, prior_mean = prior_mean, prior_cov = prior_covBAD, n_iter=1000)
    
    assert e_bad_cov.type is ValueError

bclr_good = BayesCC(X, prior_mean = prior_mean, prior_cov = prior_cov, n_iter=1000)
bclr_good.fit()

def test_cp():
    bclr_good.transform()
    assert int(bclr_good.post_k_mode) == 200
    
def test_plot():
    plot_res = bclr_good.plot_k()
    assert plot_res is None



from ..bclr_helper import std_video, gen_sim
import numpy as np

def test_vid():
    vid = np.squeeze(gen_sim(n=1, plus = -2, ind=25))
    std_vid = std_video(vid)
    vm = np.mean(std_vid, axis=(1,2))
    vsd = np.std(std_vid, axis=(1,2))
    
    np.testing.assert_almost_equal(vm, np.repeat(0, 50), 1e-10)
    np.testing.assert_almost_equal(vsd, np.repeat(1, 50), 1e-10)


from setuptools import setup, find_packages 
import codecs

DESCRIPTION = 'bclr - Bayesian changepoint detection via Logistic Regression'
with codecs.open("DESCRIPTION.md", encoding='utf-8-sig') as f:
    LONG_DESCRIPTION = f.read()
LONG_DESCRIPTION_TYPE = "text/markdown"
VERSION = {}
with open("bclr/_version.py") as fp:
    exec(fp.read(), VERSION)

setup(
    name="bclr",
    version=VERSION['__version__'],
    author="Andrew M. Thomas and Michael Jauch",
    author_email = "<me@andrewmthomas.com>",
    maintainer="Andrew M. Thomas",
    description = DESCRIPTION,
    long_description = LONG_DESCRIPTION,
    long_description_content_type = LONG_DESCRIPTION_TYPE,
    packages=find_packages(),
    include_package_data=True,
    package_data={'bclr': ['bclr/tests/*.npy']},
    install_requires=[
	'wheel',
	'matplotlib',
	'pandas',
	'numpy <= 1.26.4',
	'detectda >= 0.5.3',
	'joblib',
	'tabulate',
	'scikit-learn >= 1.3.0',
	'scipy',
	'polyagamma >= 1.3.6',
        'ruptures >= 1.1.8'
    ],
    keywords = ['changepoint', 'bayesian', 'logistic regression'],
    classifiers = [
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python'
    ]
)



import numpy as np
from scipy.stats import binom
import warnings

class SegmentationWarning(Warning):
    pass

def _proc_cov(prior_cov, p):
    """
    Process covariance matrix to ensure it is valid and convert it to np.ndarray.

    Parameters
    ----------
    prior_cov : ndarray of shape (d,d)
        Symmetric positive (semi)definite prior covariance matrix. 
    p : int
        Integer corresponding to data dimension.

    Raises
    ------
    ValueError
        If prior_cov is invalid, raise error.

    Returns
    -------
    prior_cov : np.ndarray of shape (d,d)
        Symmetric positive (semi)definite prior covariance matrix. 

    """
    prior_cov = np.array(prior_cov)
    dim = prior_cov.shape[0]
    if dim != p:
        msg_dim = (f"Data is {p}-dimensional but prior covariance matrix is {dim}x{dim}."
                       f" Please specify a {p}x{p} prior covariance matrix")
        raise ValueError(msg_dim)
    
    msg_sym = "Prior covariance matrix should be symmetric"
    if dim != prior_cov.shape[1]:
        raise ValueError(msg_sym)
    elif not np.all(np.isclose(prior_cov, prior_cov.T)):
        raise ValueError(msg_sym)
    
    if np.any(np.linalg.eigvals(prior_cov) < 0):
        raise ValueError("Prior covariance matrix must be positive semi-definite")
    
    return prior_cov

def _proc_mean_cov(prior_mean, prior_cov, p):
    """
    Process mean and covariance to ensure they are valid and convert them to np.ndarray.

    Parameters
    ----------
    prior_mean : ndarray of length d
        Array containing prior mean of Normal distribution.
    prior_cov : ndarray of shape (d,d)
        Symmetric positive (semi)definite prior covariance matrix. 
    p : int
        Integer corresponding to dimension of data.

    Raises
    ------
    ValueError
        If prior_mean and prior_cov are invalid, raise error.

    Returns
    -------
    prior_mean : np.ndarray of length d
        Array containing prior mean of Normal distribution.
    prior_cov : np.ndarray of shape (d,d)
        Symmetric positive (semi)definite prior covariance matrix. 

    """
    prior_mean = np.array(prior_mean)
    prior_cov = _proc_cov(prior_cov, p)
    
    l = len(prior_mean)
    if l == p:
        prior_mean = np.array(prior_mean)
        if len(prior_mean.shape) > 1:
            raise ValueError("Prior mean vector should be one-dimensional")
    else:
        msg_len = (f"Prior mean has length {l} but should have length {p} based on the dimension of the data")
        raise ValueError(msg_len)
        
    return prior_mean, prior_cov
        
    
def uni_binom(n, p, lam):
    """
    Interpolates between a discrete uniform and a unit-shifted binomial distribution .

    Parameters
    ----------
    n : int
        Number of trials parameter for binomial distribution.
    p : float
        Parameter for binomial distribution on [0, n-1].
    lam : float
        Interpolation parameter. Between 0 and 1.

    Returns
    -------
    ndarray
        Probability mass function on [0, n-1] (equivalently [1,n]).

    """
    if lam < 0 or lam > 1:
        raise ValueError("Lambda must be nonnegative and no greater than 1")
    k = np.arange(1,n+1)
    with warnings.catch_warnings():
        # suppress the RuntimeWarning that sometimes results...
        warnings.simplefilter("ignore")
        bin_p = binom.pmf(k-1, n-1, p)
    return bin_p**(lam) 

def prob_mode(arr):
    """
    Gets mode of given array. Gives error based on ``bclr_multi`` if
    min_size constraints yield an empty array which is fed to this function.

    Parameters
    ----------
    arr : ndarray of shape (n_obs,)
        Array of observed numbers (need not be integers).

    Returns
    -------
    ndarray
        Returns array of mode(s)..

    """
    arr_vals, arr_counts = np.unique(arr, return_counts=True)
    if len(arr_counts)==0:
        warnings.warn_explicit(message="""Argument min_size too large to find posterior mode... \n""", category=SegmentationWarning, 
                               filename="bclr_multi.py", lineno=116)
        return np.nan
    else:
        return arr_vals[np.argmax(arr_counts)]

def std_video(video, flip=False):
    """
    A function to standard a video to have frames with mean 0, variance 1 pixel intensity.

    Parameters
    ----------
    video : array-like of shape (n_frames, n_rows, n_cols)
        The greyscale video of interest
    flip : bool, optional
        If True, pixels in image will be inverted. The default is False.

    Returns
    -------
    ndarray
        Standardized video.

    """
    v_mean = np.mean(video, axis=(1,2))
    v_std = np.std(video, axis=(1,2))
    v_means=np.transpose(np.tile(v_mean, (video.shape[1], video.shape[2],1)), (2,0,1))
    v_stds=np.transpose(np.tile(v_std, (video.shape[1], video.shape[2],1)), (2,0,1))
    return (-1)**(flip)*(video-v_means)/(v_stds)

def gen_sim(n=1000, plus = -2, ind=25, seed=0):
    """
    The function used to generate the simulated videos in Thomas, Jauch, and Matteson (2023).

    Parameters
    ----------
    n : int, optional
        The number of simulated videos to generate. The default is 1000.
    plus : bool, optional
        Add this quantity to the random rectangular region within each image. The default is -2.
    ind : int, optional
        Where to locate the changepoint. The default is 25.
    seed : int, optional
        Seed for random number generation. The default is 0.

    Returns
    -------
    arr : array of shape (n, 50, 50, 50)
        Contains all simulated videos

    """
    np.random.seed(seed)
    k = 0
    arr = np.empty((n, 50, 50, 50))
    while k < n:
        noise = np.random.randn(50, 50, 50)
        for i in range(ind, 50):
            pt = np.random.randint(-20, 20, 2)
            size = np.random.randint(2, 5, 2)
            y1, y2, z1, z2 = (25-size[0]+pt[0]), (25+size[0]+pt[0]), (25-size[1]+pt[1]), (25+size[1]+pt[1])
            noise[i, y1:y2, z1:z2] = noise[i, y1:y2, z1:z2]+plus
    
        arr[k] = std_video(noise)
        k += 1
    return arr
